from abc import ABC, abstractmethod

class BaseTrainer(ABC):
    def __init__(self, config=None):
        self.config = config
        self._rabbitmq_config = config or {}
        self.rabbitmq_connection = None
        self.rabbitmq_channel = None

    @property
    def rabbitmq_config(self):
        return self._rabbitmq_config

    @abstractmethod
    def train(self, data, labels):
        pass

    @abstractmethod
    def evaluate(self, test_data):
        pass

    @abstractmethod
    def save_model(self, file_path):
        pass

    @abstractmethod
    def load_model(self, file_path):
        pass


from nedo_vision_training.client.GrpcClientBase import GrpcClientBase
from nedo_vision_training.protos.TrainingAgentService_pb2 import UpdateTrainingAgentStatusRequest
from nedo_vision_training.protos.TrainingAgentService_pb2_grpc import TrainingAgentServiceStub
from nedo_vision_training.exceptions import GrpcClientError
import time
import logging

logger = logging.getLogger(__name__)

class TrainingAgentStatusClient(GrpcClientBase):
    def __init__(self, host: str, port: int, token: str):
        """
        Initialize the TrainingAgentStatusClient for updating trainingagent status.
        
        Args:
            host (str): The gRPC server host.
            port (int): The gRPC server port.
            token (str): Authentication token for the training agent.
        """
        super().__init__(host, port)
        self.token = token
        self.connect(TrainingAgentServiceStub)

    def update_status(self, status_code: str) -> bool:
        """
        Update training agent status
        
        Args:
            status_code: Status code to update to
            
        Returns:
            bool: True if successful, False otherwise
        """
        try:
            if not self.stub:
                raise GrpcClientError("Not connected to manager")

            # Create request
            request = UpdateTrainingAgentStatusRequest(
                token=self.token,
                status_code=status_code,
                timestamp=int(time.time() * 1000)  # Current timestamp in milliseconds
            )

            # Send request using base class error handling
            response = self.handle_rpc(self.stub.UpdateStatus, request)
            return response.success if response else False

        except Exception as e:
            logger.error(f"Failed to update status: {e}")
            return False


import time
import json
import os
import tempfile
import zipfile
import shutil
import requests
from typing import List, Dict, Any
from nedo_vision_training.modules.dataset.DatasetHandler import DatasetHandler
from nedo_vision_training.logger.Logger import Logger

logger = Logger(__name__)

class COCODatasetHandler(DatasetHandler):
    def __init__(self, dataset_id: str, dataset_snapshot_id: str, output_dir: str, 
                 train_ratio: float = 0.7, 
                 val_ratio: float = 0.15, 
                 test_ratio: float = 0.15,
                 seed: int = 42,
                 config: dict = None,
                 skip_label_transform: bool = False):
        super().__init__(dataset_id, output_dir, train_ratio, val_ratio, test_ratio, seed, config)
        self.dataset_snapshot_id = dataset_snapshot_id
        self.val_dir = os.path.join(output_dir, "valid")
        
        if not config:
            raise ValueError("Config is required for COCODatasetHandler")
            
        self.server_host = config.get('server_host', 'localhost')
        self.rest_api_port = config.get('rest_api_port', 8081)
        self.token = config.get('token')

        self.skip_label_transform = skip_label_transform
        
        self.api_base_url = f"http://{self.server_host}:{self.rest_api_port}/api/v1"
        
        if not self.token:
            raise ValueError("Training agent token is required. Ensure 'token' is set in config.")
    
    def _download_dataset_from_api(self, temp_dir: str) -> str:
        if not self.token:
            raise Exception("Training agent token is required")

        headers = {
            "X-Training-Agent-Token": self.token
        }

        base_url = f"{self.api_base_url}/dataset/{self.dataset_id}/snapshot/{self.dataset_snapshot_id}/training-agent"

        def get_download_url() -> str:
            url = f"{base_url}/generate-download-url?format=coco"
            res = requests.get(url, headers=headers, timeout=60)
            res.raise_for_status()
            data = res.json()
            return data.get("data", {}).get("url", "")

        def trigger_generate():
            url = f"{base_url}/generate-format?format=coco"
            logger.info("Triggering dataset format generation...")
            res = requests.get(url, headers=headers, timeout=60)
            res.raise_for_status()

        def get_status():
            url = f"{base_url}/generate-status?format=coco"
            res = requests.get(url, headers=headers, timeout=60)
            res.raise_for_status()
            return res.json().get("data", {})

        def download_file(download_url: str) -> str:
            logger.info("Downloading dataset from presigned URL")

            response = requests.get(download_url, stream=True, timeout=300)
            response.raise_for_status()

            total_size = int(response.headers.get("content-length", 0))
            downloaded = 0
            next_log_percent = 0

            timestamp = int(time.time() * 1000)
            process_id = os.getpid()
            zip_file_path = os.path.join(
                temp_dir,
                f"dataset_{self.dataset_id}_{timestamp}_{process_id}.zip"
            )

            with open(zip_file_path, "wb") as f:
                for chunk in response.iter_content(chunk_size=8192):
                    if not chunk:
                        continue

                    f.write(chunk)
                    downloaded += len(chunk)

                    if total_size > 0:
                        percent = (downloaded / total_size) * 100
                        if percent >= next_log_percent:
                            logger.info(f"Downloading... {percent:.2f}%")
                            next_log_percent += 10

            logger.info(f"Download completed: {zip_file_path}")
            return zip_file_path

        try:
            # STEP 1: Try get existing URL
            logger.info("Checking if dataset already generated...")
            download_url = get_download_url()

            if download_url:
                logger.info("Found existing download URL, skipping generation")
                return download_file(download_url)

            # STEP 2: Trigger generation
            trigger_generate()

            # STEP 3: Polling
            logger.info("Polling dataset generation status...")
            max_wait_time = 60 * 30  # 30 minutes max
            interval = 20
            waited = 0

            while waited < max_wait_time:
                status_data = get_status()

                status = status_data.get("status")
                progress = status_data.get("generateProgress", 0)
                retry = status_data.get("retryCount", 0)
                fail_reason = status_data.get("failReason", "")

                logger.info(
                    f"Status: {status} | Progress: {progress:.2f}% | Retry: {retry}"
                )

                if status == "Generated":
                    logger.info("Dataset generation completed")
                    break

                if status == "Failed":
                    raise Exception(f"Dataset generation failed: {fail_reason}")

                time.sleep(interval)
                waited += interval

            else:
                raise Exception("Timeout waiting for dataset generation")

            # STEP 4: Get URL again
            download_url = get_download_url()

            if not download_url:
                raise Exception("Generated but download URL is empty")

            return download_file(download_url)

        except Exception as e:
            logger.error(f"Failed to download dataset via API: {e}")
            raise

    
    def _extract_dataset_zip(self, zip_file_path: str, extract_dir: str) -> Dict[str, Any]:
        """
        Extract dataset zip file and organize files by split
        
        Args:
            zip_file_path (str): Path to the zip file
            extract_dir (str): Directory to extract files to
            
        Returns:
            Dict[str, Any]: Information about extracted dataset including COCO annotations
        """
        try:
            logger.info(f"Extracting dataset zip file: {zip_file_path} to {extract_dir}")
            
            with zipfile.ZipFile(zip_file_path, 'r') as zip_ref:
                zip_ref.extractall(extract_dir)

            # Search for annotation file
            coco_annotations_path = os.path.join(extract_dir, "_annotations.coco.json")
            
            if coco_annotations_path and os.path.exists(coco_annotations_path):
                logger.info(f"Found COCO annotations file: {coco_annotations_path}")
                with open(coco_annotations_path, 'r') as f:
                    coco_data = json.load(f)
            else:
                logger.warning("No COCO annotations file found in the extracted dataset")
                coco_data = None
            
            image_files = []
            # Check if images are in a subdirectory
            images_dir = os.path.join(extract_dir, "images")
            search_dir = images_dir if os.path.exists(images_dir) and os.path.isdir(images_dir) else extract_dir
            
            logger.info(f"Searching for images in: {search_dir}")
            
            for file in os.listdir(search_dir):
                if file.lower().endswith(('.jpg', '.jpeg', '.png', '.bmp', '.tiff', '.webp')):
                    image_files.append(os.path.join(search_dir, file))
            
            logger.info(f"Extracted {len(image_files)} image files from dataset")
            
            return {
                'extract_dir': extract_dir,
                'coco_data': coco_data,
                'image_files': image_files,
                'coco_annotations_path': coco_annotations_path
            }
            
        except Exception as e:
            logger.error(f"Failed to extract dataset zip: {e}")
            raise Exception(f"Failed to extract dataset: {e}")
        
    def process_dataset(self) -> Dict[str, Any]:
        """
        Process the dataset by downloading from manager API with COCO annotations and organizing into train/val/test splits
        
        Returns:
            Dict[str, Any]: Processing results including success status and metadata
        """
        try:
            self._ensure_output_dir()
            os.makedirs(self.val_dir, exist_ok=True)
            
            with tempfile.TemporaryDirectory() as temp_dir:
                logger.info(f"Processing dataset {self.dataset_id} using API download")
                
                zip_file_path = self._download_dataset_from_api(temp_dir)
                extraction_result = self._extract_dataset_zip(zip_file_path, os.path.join(temp_dir, "extracted"))
                
                coco_data = extraction_result['coco_data']
                image_files = extraction_result['image_files']
                extract_dir = extraction_result['extract_dir']
                
                if not coco_data:
                    logger.warning("No COCO annotations found, creating splits without annotations")
                    return self._process_without_annotations(image_files)
                
                return self._process_with_coco_annotations(coco_data, image_files, extract_dir, self.skip_label_transform)
                
        except Exception as e:
            logger.error(f"Failed to process dataset: {e}")
            return {
                "success": False,
                "error": str(e),
                "message": f"Failed to process dataset: {e}"
            }
    
    def _process_with_coco_annotations(
        self, 
        coco_data: Dict[str, Any], 
        image_files: List[str], 
        extract_dir: str,
        skip_label_transform: bool = False  # ← new parameter
    ) -> Dict[str, Any]:
        """
        Process dataset with existing COCO annotations and create train/val/test splits
        
        Args:
            coco_data (Dict[str, Any]): COCO format annotations
            image_files (List[str]): List of image file paths
            extract_dir (str): Directory where images were extracted
            skip_label_transform (bool): If True, skip category ID remapping to 0-based.
                                        Use this when the downstream trainer (e.g. RF-DETR)
                                        handles its own ID convention internally.
                
        Returns:
            Dict[str, Any]: Processing results
        """
        try:
            logger.info("Processing dataset with existing COCO annotations")
            
            # Create mapping from image filename to full path
            image_path_map = {os.path.basename(img_path): img_path for img_path in image_files}
            
            # Get images and annotations from COCO data
            images = coco_data.get('images', [])
            annotations = coco_data.get('annotations', [])
            categories = coco_data.get('categories', [])
            
            if skip_label_transform:
                # Skip ID remapping — keep original category IDs as-is
                logger.info(f"⏭️ Skipping label transformation. Using original category IDs: "
                            f"{[c['id'] for c in categories]}")
                converted_categories = [c.copy() for c in categories]
                converted_annotations = [ann.copy() for ann in annotations]
            else:
                # Remap category IDs to 0-based indexing
                category_id_mapping = {}
                converted_categories = []
                for i, category in enumerate(categories):
                    old_id = category['id']
                    new_id = i
                    category_id_mapping[old_id] = new_id
                    
                    converted_category = category.copy()
                    converted_category['id'] = new_id
                    converted_categories.append(converted_category)
                
                logger.info(f"Converted {len(categories)} categories from 1-based to 0-based indexing: "
                            f"{category_id_mapping}")
                
                converted_annotations = []
                for ann in annotations:
                    converted_ann = ann.copy()
                    old_category_id = ann.get('category_id')
                    if old_category_id in category_id_mapping:
                        converted_ann['category_id'] = category_id_mapping[old_category_id]
                        converted_annotations.append(converted_ann)
                    else:
                        logger.warning(f"Annotation has unknown category_id {old_category_id}, skipping")

            # Create image_id to annotations mapping
            image_annotations = {}
            for ann in converted_annotations:
                image_id = ann['image_id']
                if image_id not in image_annotations:
                    image_annotations[image_id] = []
                image_annotations[image_id].append(ann)
            
            # Split images into train/val/test
            images_copy = images.copy()
            train_images, val_images, test_images = self._split_dataset(images_copy)
            
            # Process each split
            splits = {
                "train": (train_images, self.train_dir),
                "val": (val_images, self.val_dir),
                "test": (test_images, self.test_dir)
            }
            
            all_split_data = {}
            
            for split_name, (split_images, split_dir) in splits.items():
                logger.info(f"Processing {split_name} split with {len(split_images)} images")
                
                split_coco_data = {
                    "images": [],
                    "annotations": [],
                    "categories": converted_categories,
                    "info": {
                        "description": f"{split_name} split of dataset {self.dataset_id}",
                        "version": "1.0",
                        "year": 2024
                    }
                }
                
                for img in split_images:
                    img_filename = img['file_name']
                    if img_filename in image_path_map:
                        src_path = image_path_map[img_filename]
                        dst_path = os.path.join(split_dir, img_filename)
                        
                        try:
                            shutil.copy2(src_path, dst_path)
                            split_coco_data["images"].append(img)
                            
                            if img['id'] in image_annotations:
                                split_coco_data["annotations"].extend(image_annotations[img['id']])
                                
                        except Exception as e:
                            logger.error(f"Failed to copy image {img_filename}: {e}")
                            continue
                    else:
                        logger.warning(f"Image file not found: {img_filename}")
                
                split_file = os.path.join(split_dir, "_annotations.coco.json")
                with open(split_file, 'w') as f:
                    json.dump(split_coco_data, f, indent=2)
                
                all_split_data[split_name] = split_coco_data
                logger.info(f"Created {split_name} split with {len(split_coco_data['images'])} images "
                            f"and {len(split_coco_data['annotations'])} annotations")
            
            logger.info("Dataset successfully processed with COCO annotations and split into train/val/test")
            
            return {
                "success": True,
                "message": "Dataset successfully processed with COCO annotations and split into train/val/test",
                "output_dir": self.output_dir,
                "categories": converted_categories,
                "splits": {
                    "train": len(all_split_data["train"]["images"]),
                    "val": len(all_split_data["val"]["images"]),
                    "test": len(all_split_data["test"]["images"])
                }
            }
            
        except Exception as e:
            logger.error(f"Failed to process dataset with COCO annotations: {e}")
            raise
    
    def _process_without_annotations(self, image_files: List[str]) -> Dict[str, Any]:
        """
        Process dataset without annotations (images only) and create splits
        
        Args:
            image_files (List[str]): List of image file paths
            
        Returns:
            Dict[str, Any]: Processing results
        """
        try:
            logger.info("Processing dataset without annotations (images only)")
            
            # Split images into train/val/test
            train_images, val_images, test_images = self._split_dataset(image_files)
            
            # Process each split
            splits = {
                "train": (train_images, self.train_dir),
                "val": (val_images, self.val_dir), 
                "test": (test_images, self.test_dir)
            }
            
            for split_name, (split_image_files, split_dir) in splits.items():
                logger.info(f"Processing {split_name} split with {len(split_image_files)} images")
                
                # Copy images to split directory
                for img_path in split_image_files:
                    try:
                        dst_path = os.path.join(split_dir, os.path.basename(img_path))
                        shutil.copy2(img_path, dst_path)
                    except Exception as e:
                        logger.error(f"Failed to copy image {img_path}: {e}")
                        continue
                
                # Create empty COCO annotations file
                split_coco_data = {
                    "images": [],
                    "annotations": [],
                    "categories": [],
                    "info": {
                        "description": f"{split_name} split of dataset {self.dataset_id} (no annotations)",
                        "version": "1.0",
                        "year": 2024
                    }
                }
                
                split_file = os.path.join(split_dir, "_annotations.coco.json")
                with open(split_file, 'w') as f:
                    json.dump(split_coco_data, f, indent=2)
            
            logger.info("Dataset successfully processed without annotations and split into train/val/test")
            
            return {
                "success": True,
                "message": "Dataset successfully processed without annotations and split into train/val/test",
                "output_dir": self.output_dir,
                "categories": [],
                "splits": {
                    "train": len(train_images),
                    "val": len(val_images),
                    "test": len(test_images)
                }
            }
            
        except Exception as e:
            logger.error(f"Failed to process dataset without annotations: {e}")
            raise
            
    def _convert_annotations(self, annotations: List[Any]) -> List[Any]:
        """
        Convert annotations to the required format (abstract method implementation)
        
        Args:
            annotations (List[Any]): Original annotations
            
        Returns:
            List[Any]: Converted annotations (pass-through in this implementation)
        """
        # Required by base class but not used since we handle COCO data directly
        return annotations 


import logging
import grpc
from nedo_vision_training.protos.DatasetService_pb2 import GetDatasetRequest
from nedo_vision_training.protos.DatasetService_pb2_grpc import DatasetServiceStub
from nedo_vision_training.client.GrpcClientBase import GrpcClientBase

logger = logging.getLogger(__name__)

class DatasetServiceClient(GrpcClientBase):
    def __init__(self, server_host: str, server_port: int):
        """
        Initialize the DatasetServiceClient for fetching dataset information.
        Args:
            server_host (str): The gRPC server host.
            server_port (int): The gRPC server port.
        """
        super().__init__(server_host, server_port)
        try:
            self.connect(DatasetServiceStub)
        except Exception as e:
            logging.error(f"Failed to connect to gRPC server: {e}")
            self.stub = None
            
    def get_dataset(self, dataset_id: str) -> dict:
        """
        Fetch dataset information from the server.
        
        Args:
            dataset_id (str): The ID of the dataset to fetch.
            
        Returns:
            dict: Result of the dataset fetch operation containing:
                - success (bool): Whether the operation was successful
                - message (str): Additional information
                - data (list): List of dataset items if successful
        """
        if not self.stub:
            return {"success": False, "message": "gRPC connection is not established."}
        
        try:
            request = GetDatasetRequest(dataset_id=dataset_id)
            response = self.handle_rpc(self.stub.GetDataset, request)
            
            if response and response.success:
                return {
                    "success": True,
                    "message": response.message,
                    "data": response.data
                }
                
            return {
                "success": False,
                "message": response.message if response else "Unknown error",
                "data": None
            }
            
        except grpc.RpcError as e:
            logger.error(f"gRPC error while fetching dataset {dataset_id}: {str(e)}")
            return {"success": False, "message": f"RPC error: {str(e)}", "data": None}
            
        except Exception as e:
            logger.error(f"Unexpected error while fetching dataset {dataset_id}: {str(e)}")
            return {"success": False, "message": f"An unexpected error occurred: {str(e)}", "data": None} 


import logging
import re
import uuid
from nedo_vision_training.logger.Logger import Logger
from nedo_vision_training.utils.platform_detector import PlatformDetector
from nedo_vision_training.utils.networking import Networking

logger = Logger()

class AppInitializer:
    @staticmethod
    def validate_uuid(value):
        """Validate if the provided value is a valid UUID."""
        try:
            uuid.UUID(value)
            return value
        except ValueError:
            raise ValueError(f"Invalid device ID format: {value}. Must be a valid UUID.")

    @staticmethod
    def validate_server_host(value):
        """Validate if the server host is a valid domain name or IP address."""
        domain_regex = (
            r"^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\-]*[a-zA-Z0-9])\.)*"
            r"([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\-]*[A-Za-z0-9])$"
        )
        ip_regex = (
            r"^(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
            r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
            r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\."
            r"(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)$"
        )
        if re.match(domain_regex, value) or re.match(ip_regex, value):
            return value
        raise ValueError(f"Invalid server host: {value}. Must be a valid domain or IP address.")

    def initialize(self):
        """
        Placeholder for any future initialization logic. No registration is performed.
        """
        logger.info("AppInitializer: No registration logic required. Initialization complete.")



import time
import multiprocessing
import signal
import sys

# Set multiprocessing start method to 'spawn' for CUDA compatibility
try:
    multiprocessing.set_start_method('spawn', force=True)
except RuntimeError:
    # Method already set, ignore
    pass

from nedo_vision_training.exceptions import GrpcClientError
from nedo_vision_training.client.ConnectionInfoClient import ConnectionInfoClient
from nedo_vision_training.client.SystemUsageClient import SystemUsageClient
from nedo_vision_training.client.TrainingAgentStatusClient import TrainingAgentStatusClient
from nedo_vision_training.logger.Logger import Logger
from nedo_vision_training.services.WorkerManager import WorkerManager


class TrainingService:
    """
    Main training service class that manages the training agent lifecycle.
    Uses token-based authentication only. No config file.
    """
    
    def __init__(
        self,
        token: str,
        server_host: str = "localhost",
        server_port: int = 50051,
        rest_api_port: int = 8081,
        system_usage_interval: int = 5,
        latency_check_interval: int = 10,
    ):
        """
        Initialize the training service.
        
        Args:
            token: Training agent token (required)
            server_host: Manager server host (default: 'localhost')
            server_port: Manager server gRPC port (default: 50051)
            rest_api_port: Manager REST API port (default: 8081)
            system_usage_interval: Interval for system usage reporting in seconds (default: 5)
            latency_check_interval: Interval for latency monitoring in seconds (default: 10)
        """
        self.logger = Logger()
        self.connection_info_client = None
        self.system_usage_client = None
        self.status_client = None
        self.worker_manager = None
        self.running = False
        self.token = token
        self.manager_host = server_host
        self.manager_port = server_port
        self.rest_api_port = rest_api_port
        self.system_usage_interval = system_usage_interval
        self.latency_check_interval = latency_check_interval
        
        # Setup signal handlers
        signal.signal(signal.SIGINT, self._signal_handler)
        signal.signal(signal.SIGTERM, self._signal_handler)

    def initialize(self) -> bool:
        """
        Initialize the training service components.
        
        Returns:
            bool: True if initialization successful, False otherwise
        """
        try:
            self.logger.info("Training service initialization started")
            if not self.token:
                raise ValueError("Training agent token is required")

            # Fetch configuration from manager via gRPC
            self.connection_info_client = ConnectionInfoClient(
                host=self.manager_host,
                port=self.manager_port,
                token=self.token
            )
            connection_info = self.connection_info_client.get_connection_info()
            if not connection_info.get("success"):
                raise RuntimeError(f"Failed to fetch connection info: {connection_info}")
            self.config = connection_info
            self.config["server_host"] = self.manager_host
            self.config["server_port"] = self.manager_port
            self.config["rest_api_port"] = self.rest_api_port
            self.config["token"] = self.token
            self.config["system_usage_interval"] = self.system_usage_interval
            self.config["latency_check_interval"] = self.latency_check_interval

            # Initialize gRPC clients
            self._initialize_grpc_clients()

            # Initialize WorkerManager
            self.worker_manager = WorkerManager(self.config)

            self.logger.info("Training service initialization completed successfully")
            return True
        except Exception as e:
            self.logger.error(f"Failed to initialize training service: {e}")
            import traceback
            self.logger.error(traceback.format_exc())
            return False

    def _initialize_grpc_clients(self):
        """Initialize gRPC clients for communication with the manager"""
        try:
            # Initialize connection info client (already done in initialize)
            # Initialize system usage client
            self.system_usage_client = SystemUsageClient(
                host=self.manager_host,
                port=self.manager_port,
                token=self.token
            )
            # Initialize status client
            self.status_client = TrainingAgentStatusClient(
                host=self.manager_host,
                port=self.manager_port,
                token=self.token
            )

            self.logger.info("gRPC clients initialized and connected successfully")
        except Exception as e:
            import traceback
            self.logger.error(f"Failed to initialize gRPC clients: {e}")
            self.logger.error(traceback.format_exc())
            raise GrpcClientError(f"Failed to initialize gRPC clients: {e}")
    
    def start(self):
        """Start the training service"""
        if not self.running:
            self.running = True
            self.logger.info("Training service started")
            try:
                # Update status to connected
                self.status_client.update_status("connected")
                # Start all workers via WorkerManager
                self.worker_manager.start_all()
                # Block main thread to keep process alive
                while self.running:
                    time.sleep(1)
            except Exception as e:
                self.logger.error(f"Error in training service: {e}")
                import traceback
                self.logger.error(traceback.format_exc())
                self.stop()
        else:
            self.logger.info("Service already running.")
    
    def stop(self):
        """Stop the training service"""
        if self.running:
            self.running = False
            self.logger.info("Training service stopping...")
            try:
                # Update status to disconnected
                if self.status_client:
                    self.status_client.update_status("disconnected")
                # Stop all workers via WorkerManager
                if hasattr(self, 'worker_manager'):
                    self.worker_manager.stop_all()
                # Disconnect gRPC clients
                if self.connection_info_client:
                    self.connection_info_client.close()
                if self.system_usage_client:
                    self.system_usage_client.close()
                if self.status_client:
                    self.status_client.close()
                self.logger.info("Training service stopped")
            except Exception as e:
                self.logger.error(f"Error stopping training service: {e}")
                import traceback
                self.logger.error(traceback.format_exc())
        else:
            self.logger.info("Service already stopped.")
    
    def _signal_handler(self, signum, frame):
        """Handle system signals for graceful shutdown"""
        self.logger.info(f"Received signal {signum}, shutting down...")
        self.stop()
        sys.exit(0)
    
    def run(self):
        """Run the training service"""
        if self.initialize():
            self.start()
            sys.exit(1)


def main():
    """Main entry point for the training service"""
    import argparse
    
    parser = argparse.ArgumentParser(description="Nedo Vision Training Service")
    parser.add_argument(
        "--token", 
        required=True,
        help="Training agent token (required)"
    )
    parser.add_argument(
        "--server-host", 
        required=True,
        help="Manager server host (required)"
    )
    parser.add_argument(
        "--server-port", 
        type=int,
        required=True,
        help="Manager server gRPC port (required)"
    )
    parser.add_argument(
        "--rest-api-port", 
        type=int,
        default=8081,
        help="Manager REST API port (default: 8081)"
    )
    parser.add_argument(
        "--system-usage-interval",
        type=int,
        default=5,
        help="System usage reporting interval in seconds (default: 5)"
    )
    parser.add_argument(
        "--latency-check-interval",
        type=int,
        default=10,
        help="Latency monitoring interval in seconds (default: 10)"
    )
    args = parser.parse_args()
    
    service = TrainingService(
        token=args.token,
        server_host=args.server_host,
        server_port=args.server_port,
        rest_api_port=args.rest_api_port,
        system_usage_interval=args.system_usage_interval,
        latency_check_interval=args.latency_check_interval
    )
    service.run()


if __name__ == "__main__":
    main() 


import threading
import time
from nedo_vision_training.client.GrpcClientBase import GrpcClientBase
from nedo_vision_training.client.SystemUsageClient import SystemUsageClient
from nedo_vision_training.logger.Logger import Logger
from nedo_vision_training.utils.system_monitor import SystemMonitor
from nedo_vision_training.utils.networking import Networking
# from client.SystemUsageClient import SystemUsageClient

logger = Logger(__name__)

class SystemUsageManager:
    def __init__(self, server_host: str, port: int, auth_token: str, latency_interval: int = 10):
        """
        Handles system usage monitoring, latency tracking, and reporting.

        Args:
            server_host (str): The gRPC server host.
            port (int): The gRPC server port.
            auth_token (str): Authentication token for the training agent.
            latency_interval (int): Interval in seconds for latency monitoring (default: 10).
        """
        if not auth_token:
            raise ValueError("⚠️ 'auth_token' cannot be empty.")

        self.system_monitor = SystemMonitor()
        self.system_usage_client = SystemUsageClient(server_host, port, auth_token)
        self.server_host = server_host
        self.port = port
        self.auth_token = auth_token
        self.latency_interval = latency_interval
        self.latency = None
        self.latency_lock = threading.Lock()
        self.stop_event = threading.Event()
        self.latency_thread = None  

        self._start_latency_monitoring()

    def _start_latency_monitoring(self):
        """Starts a background thread to monitor network latency."""
        if self.latency_thread and self.latency_thread.is_alive():
            logger.warning("⚠️ Latency monitoring thread is already running.")
            return

        self.latency_thread = threading.Thread(target=self._monitor_latency, daemon=True)
        self.latency_thread.start()
        logger.info("📡 Latency monitoring started.")

    def _monitor_latency(self):
        """Periodically checks the network latency using gRPC and updates the latency variable."""
        server_port = 50051  # Default gRPC port

        while not self.stop_event.is_set():
            try:
                latency_value = Networking.check_grpc_latency(self.server_host, server_port)
                with self.latency_lock:
                    self.latency = latency_value
                # logger.info(f"🔄 Updated network latency: {latency_value} ms")

            except Exception as e:
                logger.error("🚨 Error checking gRPC latency.", exc_info=True)
                with self.latency_lock:
                    self.latency = None  
            time.sleep(self.latency_interval) 

    def process_system_usage(self):
        """Collect and send system usage data to the server, including network latency."""
        try:
            usage = self.system_monitor.get_system_usage()
            cpu_usage = usage["cpu"]["usage_percent"]
            ram_usage = usage["ram"]
            gpu_usage = usage.get("gpu", [])

            with self.latency_lock:
                latency = self.latency if self.latency is not None else -1

            response = self.system_usage_client.send_system_usage(
                cpu_usage=cpu_usage,
                ram_usage=ram_usage,
                gpu_usage=gpu_usage,
                latency=latency,
            )

            error_message = GrpcClientBase.get_error_message(response)
            if error_message:
                logger.error(f"❌ Failed to send system usage: {error_message}")
            #else:
            #    logger.info("✅ System usage sent successfully.")

        except Exception as e:
            print(e)
            logger.error("🚨 Error sending system usage.")

    def close(self):
        """Closes the system usage client and stops the latency thread."""
        self.stop_event.set()

        if self.latency_thread and self.latency_thread.is_alive():
            self.latency_thread.join()
            logger.info("🔌 Latency monitoring thread stopped.")

        if self.system_usage_client:
            self.system_usage_client.close_client()
            logger.info("✅ SystemUsageClient closed.")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: nedo_vision_training/protos/TrainingAgentService.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'nedo_vision_training/protos/TrainingAgentService.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n6nedo_vision_training/protos/TrainingAgentService.proto\x12\x1eSindika.AspNet.App005.Services\"\x86\x02\n\x1fTrainingAgentSystemUsageRequest\x12\r\n\x05token\x18\x01 \x01(\t\x12\x11\n\tcpu_usage\x18\x02 \x01(\x02\x12\x17\n\x0f\x63pu_temperature\x18\x03 \x01(\x02\x12\x19\n\x11ram_usage_percent\x18\x04 \x01(\x02\x12\x11\n\tram_total\x18\x05 \x01(\x04\x12\x10\n\x08ram_used\x18\x06 \x01(\x04\x12\x10\n\x08ram_free\x18\x07 \x01(\x04\x12\x42\n\x03gpu\x18\x08 \x03(\x0b\x32\x35.Sindika.AspNet.App005.Services.TrainingAgentGPUUsage\x12\x12\n\nlatency_ms\x18\t \x01(\x02\"\xc0\x01\n\x15TrainingAgentGPUUsage\x12\x11\n\tgpu_index\x18\x01 \x01(\x05\x12\x19\n\x11gpu_usage_percent\x18\x02 \x01(\x02\x12\x1c\n\x14memory_usage_percent\x18\x03 \x01(\x02\x12\x1b\n\x13temperature_celsius\x18\x04 \x01(\x02\x12\x14\n\x0ctotal_memory\x18\x05 \x01(\x04\x12\x13\n\x0bused_memory\x18\x06 \x01(\x04\x12\x13\n\x0b\x66ree_memory\x18\x07 \x01(\x04\"D\n TrainingAgentSystemUsageResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"Y\n UpdateTrainingAgentStatusRequest\x12\r\n\x05token\x18\x01 \x01(\t\x12\x13\n\x0bstatus_code\x18\x02 \x01(\t\x12\x11\n\ttimestamp\x18\x03 \x01(\x03\"E\n!UpdateTrainingAgentStatusResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\")\n\x18GetConnectionInfoRequest\x12\r\n\x05token\x18\x01 \x01(\t\"\xb9\x01\n\x19GetConnectionInfoResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x15\n\rrabbitmq_host\x18\x03 \x01(\t\x12\x15\n\rrabbitmq_port\x18\x04 \x01(\x05\x12\x19\n\x11rabbitmq_username\x18\x05 \x01(\t\x12\x19\n\x11rabbitmq_password\x18\x06 \x01(\t\x12\x0f\n\x02id\x18\x07 \x01(\tH\x00\x88\x01\x01\x42\x05\n\x03_id2\xce\x03\n\x14TrainingAgentService\x12\x94\x01\n\x0fSendSystemUsage\x12?.Sindika.AspNet.App005.Services.TrainingAgentSystemUsageRequest\x1a@.Sindika.AspNet.App005.Services.TrainingAgentSystemUsageResponse\x12\x93\x01\n\x0cUpdateStatus\x12@.Sindika.AspNet.App005.Services.UpdateTrainingAgentStatusRequest\x1a\x41.Sindika.AspNet.App005.Services.UpdateTrainingAgentStatusResponse\x12\x88\x01\n\x11GetConnectionInfo\x12\x38.Sindika.AspNet.App005.Services.GetConnectionInfoRequest\x1a\x39.Sindika.AspNet.App005.Services.GetConnectionInfoResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nedo_vision_training.protos.TrainingAgentService_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_TRAININGAGENTSYSTEMUSAGEREQUEST']._serialized_start=91
  _globals['_TRAININGAGENTSYSTEMUSAGEREQUEST']._serialized_end=353
  _globals['_TRAININGAGENTGPUUSAGE']._serialized_start=356
  _globals['_TRAININGAGENTGPUUSAGE']._serialized_end=548
  _globals['_TRAININGAGENTSYSTEMUSAGERESPONSE']._serialized_start=550
  _globals['_TRAININGAGENTSYSTEMUSAGERESPONSE']._serialized_end=618
  _globals['_UPDATETRAININGAGENTSTATUSREQUEST']._serialized_start=620
  _globals['_UPDATETRAININGAGENTSTATUSREQUEST']._serialized_end=709
  _globals['_UPDATETRAININGAGENTSTATUSRESPONSE']._serialized_start=711
  _globals['_UPDATETRAININGAGENTSTATUSRESPONSE']._serialized_end=780
  _globals['_GETCONNECTIONINFOREQUEST']._serialized_start=782
  _globals['_GETCONNECTIONINFOREQUEST']._serialized_end=823
  _globals['_GETCONNECTIONINFORESPONSE']._serialized_start=826
  _globals['_GETCONNECTIONINFORESPONSE']._serialized_end=1011
  _globals['_TRAININGAGENTSERVICE']._serialized_start=1014
  _globals['_TRAININGAGENTSERVICE']._serialized_end=1476
# @@protoc_insertion_point(module_scope)



class TrainParams:
    def __init__(self, id, algorithm, dataset_id, dataset_snapshot_id, epoch, batch_size, split_ratio):
        self.id = id
        self.algorithm = algorithm
        self.dataset_id = dataset_id
        self.dataset_snapshot_id = dataset_snapshot_id
        self.epoch = epoch
        self.batch_size = batch_size
        self.split_ratio = split_ratio

    def __repr__(self):
        return f"TrainParams(id={self.id}, algorithm={self.algorithm}, " \
               f"dataset_id={self.dataset_id}, dataset_snapshot_id={self.dataset_snapshot_id}, " \
               f"epoch={self.epoch}, batch_size={self.batch_size}, " \
               f"split_ratio={self.split_ratio})"


import time
import pika
import json
from nedo_vision_training.logger.Logger import Logger

class RabbitMQClient:
    def __init__(self, config=None, heartbeat=60, blocked_connection_timeout=300, name="default"):
        """
        Initializes RabbitMQClient with connection parameters.
        
        :param config: Dictionary containing RabbitMQ connection details.
        :param heartbeat: Heartbeat interval for RabbitMQ.
        :param blocked_connection_timeout: Timeout for blocked connections.
        :param name: Identifier for the RabbitMQ client instance.
        """
        self.host = config['rabbitmq_host']
        self.port = int(config['rabbitmq_port'])
        self.username = config['rabbitmq_username']
        self.password = config['rabbitmq_password']
        self.heartbeat = heartbeat
        self.blocked_connection_timeout = blocked_connection_timeout
        self.connection = None
        self.channel = None
        self.name = name
        self.logger = Logger(f"RABBITMQ:{name}")

        if not hasattr(self, 'initialized'):
            self.initialized = True

    def connect(self, max_retries=5, backoff_factor=2):
        """Attempts to establish a connection to RabbitMQ with retry logic."""
        retries = 0
        while retries < max_retries:
            try:
                self.logger.info(f"🔌 Establishing connection to RabbitMQ ({self.name})...")
                credentials = pika.PlainCredentials(self.username, self.password)
                parameters = pika.ConnectionParameters(
                    host=self.host,
                    port=self.port,
                    credentials=credentials,
                    heartbeat=self.heartbeat,
                    blocked_connection_timeout=self.blocked_connection_timeout
                )
                self.connection = pika.BlockingConnection(parameters)
                self.channel = self.connection.channel()
                self.logger.info(f"✅ Connection to RabbitMQ ({self.name}) established successfully.")
                return
            except pika.exceptions.AMQPConnectionError as e:
                retries += 1
                delay = backoff_factor ** retries
                self.logger.error(f"❌ RabbitMQ connection error: {e}. Retrying in {delay}s...")
                time.sleep(delay)
            except pika.exceptions.ProbableAuthenticationError as e:
                self.logger.error(f"🔑 Authentication failed: {e}")
                raise
            except Exception as e:
                retries += 1
                delay = backoff_factor ** retries
                self.logger.error(f"⚠️ Unexpected error while connecting to RabbitMQ: {e}. Retrying in {delay}s...")
                time.sleep(delay)

        raise ConnectionError("❌ Failed to connect to RabbitMQ after maximum retries.")

    def close(self):
        """Closes the RabbitMQ connection and channel if open."""
        try:
            if self.channel and self.channel.is_open:
                self.channel.close()
        except Exception as e:
            self.logger.warning(f"⚠️ Error closing channel: {e}")
            
        try:
            if self.connection and self.connection.is_open:
                self.connection.close()
        except Exception as e:
            self.logger.warning(f"⚠️ Error closing connection: {e}")
            
        self.logger.info(f"🔒 RabbitMQ ({self.name}) connection closed.")

    def declare_exchange(self, exchange_name, exchange_type='direct', durable=True):
        """Declares an exchange in RabbitMQ."""
        try:
            self.channel.exchange_declare(exchange=exchange_name, exchange_type=exchange_type, durable=durable)
            self.logger.info(f"✅ Exchange '{exchange_name}' declared successfully.")
        except Exception as e:
            self.logger.error(f"❌ Error declaring exchange '{exchange_name}': {e}")
            raise

    def declare_queue(self, queue_name, durable=True, auto_delete=False):
        """Declares a queue in RabbitMQ."""
        try:
            self.channel.queue_declare(queue=queue_name, durable=durable, auto_delete=auto_delete)
            self.logger.info(f"🎯 Queue '{queue_name}' declared successfully.")
        except Exception as e:
            self.logger.error(f"❌ Error declaring queue '{queue_name}': {e}")
            raise

    def bind_queue(self, exchange_name, queue_name, routing_key):
        """Binds a queue to an exchange with a routing key."""
        try:
            self.channel.queue_bind(exchange=exchange_name, queue=queue_name, routing_key=routing_key)
            self.logger.info(f"✅ Queue '{queue_name}' bound to exchange '{exchange_name}' with routing key '{routing_key}'.")
        except Exception as e:
            self.logger.error(f"❌ Error binding queue '{queue_name}' to exchange '{exchange_name}': {e}")
            raise

    def publish_message(self, exchange_name, routing_key, message):
        """Publishes a message to an exchange."""
        try:
            if isinstance(message, dict):
                message = json.dumps(message)
            self.channel.basic_publish(
                exchange=exchange_name,
                routing_key=routing_key,
                body=message
            )
            if exchange_name != 'airosentris.status':
                self.logger.info(f"📤 Message published to exchange '{exchange_name}' with routing key '{routing_key}'.")
        except pika.exceptions.StreamLostError as e:
            self.logger.error(f"⚠️ Stream lost, attempting to reconnect: {e}")
            self.connect()
            self.publish_message(exchange_name, routing_key, message)
        except Exception as e:
            self.logger.error(f"❌ Error publishing message: {e}")
            raise

    def consume_messages(self, queue_name, on_message_callback, auto_ack=False):
        """Starts consuming messages from a specified queue."""
        try:
            self.logger.info(f"📩 Listening for messages in queue '{queue_name}'. Press CTRL+C to exit.")
            self.channel.basic_consume(queue=queue_name, on_message_callback=on_message_callback, auto_ack=auto_ack)
            self.channel.start_consuming()
            self.logger.info("🛑 Message consumption stopped.")
        except pika.exceptions.AMQPConnectionError as e:
            self.logger.error(f"⚠️ Connection error during message consumption: {e}")
            self.connect()
            self.consume_messages(queue_name, on_message_callback, auto_ack)
        except pika.exceptions.StreamLostError as e:
            self.logger.error(f"⚠️ Stream lost, attempting to reconnect: {e}")
            self.connect()
            self.consume_messages(queue_name, on_message_callback, auto_ack)
        except Exception as e:
            self.logger.error(f"❌ Unexpected error during message consumption: {e}")
            raise

    def stop_consuming(self):
        """Stops message consumption."""
        if self.channel and self.channel.is_open:
            self.channel.stop_consuming()
            self.logger.info("🛑 Stopped consuming messages.")



import torch
import os
import shutil
import tempfile
from rfdetr import RFDETRBase

from nedo_vision_training.logger.Logger import Logger
from nedo_vision_training.logger.TrainerLogger import TrainerLogger
from nedo_vision_training.modules.trainer.BaseTrainer import BaseTrainer
from nedo_vision_training.modules.dataset.COCODatasetHandler import COCODatasetHandler
from nedo_vision_training.modules.algorithm.RFDETR.RFDETRMetricsCallback import RFDETRMetricsCallback
from nedo_vision_training.modules.trainer.TrainerRegistry import TrainerRegistry

class RFDETRTrainer(BaseTrainer):
    """
    Trainer class for fine-tuning and evaluating RF-DETR models for object detection.
    Handles model training, evaluation, and model uploads.
    """
    def __init__(self, config):
        super().__init__(config)
        self.logger = Logger(__name__)
        self.trainer_logger = TrainerLogger(config)
        self.model = None
        self.train_data = None
        self.test_data = None
        self.labels = None
        self.dataset_handler = None
        self.metrics_callback = None

    def init(self, training_job):
        """
        Initialize the RFDETRTrainer with dataset, labels, and training configuration.
        """
        self.training_job = training_job  # Store the training job
        self.job_id = training_job.id
        self.trainer_logger.log_status(self.job_id, "initializing")
        self.trainer_logger.log_command(self.job_id, "Initialization started.")

        # Handle the train/test split ratio from C# service
        # C# sends split_ratio as the training portion (e.g., 0.8 = 80% train, 20% test)
        # But RF-DETR needs train/val/test, so we need to create validation from the training portion
        original_train_ratio = training_job.split_ratio
        
        # If split_ratio is too high (>= 0.95), we need to ensure validation data exists
        if original_train_ratio >= 0.95:
            self.logger.warning(f"⚠️ Split ratio {original_train_ratio} is too high for RF-DETR training. Adjusting to ensure validation set.")
            # Use 80/20 split to ensure validation data
            train_ratio = 0.8
            val_ratio = 0.2
            test_ratio = 0.0
        else:
            # Take 10% from training data for validation to ensure RF-DETR can evaluate
            # Original: train_ratio = 0.8, remaining = 0.2 (intended for test)
            # Adjusted: train_ratio = 0.7, val_ratio = 0.1, test_ratio = 0.2
            validation_from_train = min(0.1, original_train_ratio * 0.125)  # Take 12.5% of training for validation, max 10%
            train_ratio = original_train_ratio - validation_from_train
            val_ratio = validation_from_train
            test_ratio = 1.0 - original_train_ratio
        
        self.logger.info(f"📊 Dataset split configuration: Train={train_ratio:.2f}, Val={val_ratio:.2f}, Test={test_ratio:.2f}")

        # Initialize dataset handler with train/test split ratios
        self.dataset_handler = COCODatasetHandler(
            dataset_id=training_job.dataset_id,
            dataset_snapshot_id=training_job.dataset_snapshot_id,
            output_dir=os.path.join("artifacts", "datasets", str(training_job.dataset_id)),
            train_ratio=train_ratio,
            val_ratio=val_ratio,
            test_ratio=test_ratio,
            config=self.config,
            skip_label_transform=True
        )
        
        # Initialize RF-DETR model
        if torch.cuda.is_available():
            self.logger.info(f"🚀 CUDA is available")
            self.logger.info(f"🖥 GPU: {torch.cuda.get_device_name(0)}")
        else:
            self.logger.warning("⚠️ CUDA is NOT available. Training will use CPU.")
        try:
            self.model = RFDETRBase()
            self.logger.info("📦 RFDETRBase initialized")
        except Exception as e:
            self.logger.warning(f"⚠️ Error during RFDETRBase initialization: {e}")
            self.model = RFDETRBase()
            self.logger.info("📦 RFDETRBase initialized (fallback after error)")
        
        # Initialize metrics callback
        self.metrics_callback = RFDETRMetricsCallback(self.trainer_logger, self.job_id)
        
        self.trainer_logger.log_command(self.job_id, "Initialization completed.")

    def train(self):
        """
        Train the RF-DETR model using the training data.
        """
        self.trainer_logger.log_status(self.job_id, "running")
        self.trainer_logger.log_command(self.job_id, "Training started.")

        try:
            # Process dataset into COCO format
            self.logger.info(f"🔍 Processing dataset for job {self.job_id}")
            dataset_info = self.dataset_handler.process_dataset()
            
            if not dataset_info.get("success", False):
                raise Exception(f"Failed to process dataset: {dataset_info.get('error')}")
            
            # ── Fix RF-DETR supercategory bug ──────────────────────────────────────
            # RF-DETR's _load_classes() filters out categories where supercategory == "none",
            # which causes num_classes=1 and an IndexError during training.
            # We patch all annotation files to use "object" instead of "none".
            self._fix_coco_supercategories(self.dataset_handler.output_dir)
            
            # Validate dataset splits
            splits = dataset_info.get("splits", {})
            train_count = splits.get("train", 0)
            val_count = splits.get("val", 0)
            
            self.logger.info(f"📊 Dataset split results: Train={train_count}, Val={val_count}")
            
            if train_count == 0:
                raise Exception("Training set is empty. Cannot proceed with training.")
            
            if val_count == 0:
                self.logger.warning("⚠️ Validation set is empty. RF-DETR requires validation data for proper training.")
                raise Exception("Validation set is empty. RF-DETR requires validation data to prevent evaluation errors.")
            
            batch_size = self.training_job.batch_size or 1  # Default to 1 if None or 0
            if batch_size <= 0:
                batch_size = 1
                self.logger.warning(f"⚠️ Invalid batch size {self.training_job.batch_size}, using default value of 1")
            
            epochs = self.training_job.epoch or 1  # Default to 1 if None or 0
            if epochs <= 0:
                epochs = 1
                self.logger.warning(f"⚠️ Invalid epoch count {self.training_job.epoch}, using default value of 1")
            
            grad_accum_steps = max(1, int(16 / batch_size))


            # Add metrics callback
            self.model.callbacks["on_fit_epoch_end"].append(self.metrics_callback.on_fit_epoch_end)
            # Train the model with callback
            self.logger.info(f"🚀 Starting training with batch_size={batch_size}, epochs={epochs}, grad_accum_steps={grad_accum_steps}")
            
            try:
                # Prepare training arguments based on validation data availability
                train_kwargs = {
                    "dataset_dir": self.dataset_handler.output_dir,
                    "epochs": epochs,
                    "batch_size": batch_size,
                    "grad_accum_steps": grad_accum_steps,
                    "lr": 1e-4,
                    "output_dir": os.path.join("artifacts", "train", self.job_id),
                    "early_stopping": True
                }
                
                self.model.train(**train_kwargs)
            except Exception as e:
                import traceback
                self.logger.error(f"❌ Training failed: {e}")
                self.logger.error(f"❌ Full stack trace:")
                self.logger.error(traceback.format_exc())
                raise
            
            self.trainer_logger.log_command(self.job_id, "Training completed.")
            self.logger.info("✅ Training completed successfully")
                
        except Exception as e:
            self.logger.error(f"Error during training: {e}")
            # Clean up artifacts on training failure
            self.logger.info(f"🧹 Cleaning up artifacts due to training failure for job {self.job_id}")
            # self.cleanup_artifacts()
            raise

    def evaluate(self):
        """
        Evaluate the model on the test dataset.
        """
        self.trainer_logger.log_status(self.job_id, "evaluating")
        self.trainer_logger.log_command(self.job_id, "Evaluation started.")

        try:
            # Check if test data is available
            if self.test_data is None:
                self.logger.warning("⚠️ No test data available for evaluation, skipping evaluation step")
                self.trainer_logger.log_command(self.job_id, "Evaluation skipped - no test data available")
                return
            
            # Evaluate the model
            try:
                metrics = self.model.evaluate(self.test_data)
            except Exception as eval_error:
                import traceback
                self.logger.error(f"❌ Model evaluation failed: {eval_error}")
                self.logger.error(f"❌ Evaluation stack trace:")
                self.logger.error(traceback.format_exc())
                raise
            
            # Log the evaluation metrics
            if metrics:
                self.trainer_logger.log_metric(self.job_id, metrics)
            
            self.logger.info(f"Evaluation results: {metrics}")
            self.trainer_logger.log_command(self.job_id, "Evaluation completed.")
        except Exception as e:
            self.logger.error(f"Error during evaluation: {e}")
            # Clean up artifacts on evaluation failure
            self.logger.info(f"🧹 Cleaning up artifacts due to evaluation failure for job {self.job_id}")
            self.cleanup_artifacts()
            raise

    def save_model(self):
        """
        Save the trained model checkpoint.
        """
        self.trainer_logger.log_status(self.job_id, "saving")
        model_dir = os.path.join("artifacts", "models", self.job_id)
        os.makedirs(model_dir, exist_ok=True)

        try:
            # The checkpoint is saved in the output_dir used during training
            train_output_dir = os.path.join("artifacts", "train", self.job_id)
            checkpoint_path = os.path.join(train_output_dir, "checkpoint_best_total.pth")
            if not os.path.exists(checkpoint_path):
                self.logger.error(f"❌ No checkpoint found at {checkpoint_path}")
                raise ValueError("No checkpoint found after training")
            # Copy checkpoint to model_dir
            shutil.copy2(checkpoint_path, os.path.join(model_dir, "checkpoint_best_total.pth"))
            self.trainer_logger.log_command(self.job_id, "Model saving completed.")
            self.logger.info(f"✅ Model checkpoint saved to {model_dir}/checkpoint_best_total.pth")
        except Exception as e:
            self.logger.error(f"Error saving model: {e}")
            self.logger.info(f"🧹 Cleaning up artifacts due to save failure for job {self.job_id}")
            self.cleanup_artifacts()
            raise

    def upload_model(self):
        """
        Upload the trained model to storage via gRPC and update model information.
        """
        self.trainer_logger.log_command(self.job_id, "Model uploading started.")
        
        with tempfile.TemporaryDirectory() as tmp_dir:
            zip_file_path = os.path.join(tmp_dir, f"{self.job_id}.zip")
            
            model_dir = os.path.join("artifacts", "models", self.job_id)
            shutil.make_archive(zip_file_path.replace(".zip", ""), 'zip', model_dir)            
            zip_file_path = zip_file_path.replace(".zip", "") + ".zip"
            self.logger.info(f"Created model archive at {zip_file_path}")
            
            try:
                # Check if zip file exists
                if not os.path.exists(zip_file_path):
                    self.logger.error(f"❌ Model archive not found at {zip_file_path}")
                    raise ValueError("Model archive not found for upload")
                
                # Upload via gRPC
                response = self.trainer_logger.upload_model(self.job_id, zip_file_path)
                
                if response.get("success"):
                    self.logger.info("Model successfully uploaded via gRPC.")
                    # Note: upload_model in manager already updates the file path, 
                    # but we can log it or double check if needed.
                    # The response contains the file_path.
                    s3_path = response.get("file_path")
                    # We don't need to call log_model_file_path again if the upload method does it,
                    # but TrainerLogger.upload_model just calls the gRPC method.
                    # The gRPC method in Manager updates the job entity.
                    # So we are good.
                else:
                    raise Exception(f"Failed to upload model: {response.get('message')}")
                
            except Exception as e:
                self.logger.error(f"Error uploading model: {e}")
                self.trainer_logger.log_command(self.job_id, f"Model upload failed: {e}")
                raise
                self.logger.error(f"An error occurred during model upload: {str(e)}")
                raise
        
        self.trainer_logger.log_command(self.job_id, "Model uploading completed.")
        self.trainer_logger.log_status(self.job_id, "completed")

    def cleanup_artifacts(self):
        """
        Clean up training artifacts for this job.
        """
        try:
            # Clean up dataset artifacts
            if self.dataset_handler and hasattr(self.dataset_handler, 'output_dir'):
                dataset_dir = self.dataset_handler.output_dir
                if os.path.exists(dataset_dir):
                    shutil.rmtree(dataset_dir)
                    self.logger.info(f"🧹 Cleaned up dataset artifacts for job {self.job_id}")
            
            # Clean up model artifacts
            model_dir = os.path.join("artifacts", "models", self.job_id)
            if os.path.exists(model_dir):
                shutil.rmtree(model_dir)
                self.logger.info(f"🧹 Cleaned up model artifacts for job {self.job_id}")
            
            # Clean up training artifacts
            train_dir = os.path.join("artifacts", "train", self.job_id)
            if os.path.exists(train_dir):
                shutil.rmtree(train_dir)
                self.logger.info(f"🧹 Cleaned up training artifacts for job {self.job_id}")
                
        except Exception as e:
            self.logger.error(f"Error cleaning up artifacts for job {self.job_id}: {e}")

    def load_model(self, file_path):
        """
        Load a trained RF-DETR model from checkpoint.
        """
        self.trainer_logger.log_command(self.job_id, f"Model loading started from {file_path}.")
        try:
            self.model = RFDETRBase(pretrain_weights=file_path)
            self.trainer_logger.log_command(self.job_id, "Model loading completed.")
        except Exception as e:
            self.logger.error(f"Error loading model: {e}")
            raise 

    def _fix_coco_supercategories(self, dataset_dir: str):
        """
        RF-DETR's _load_classes() filters out categories where supercategory == "none",
        causing num_classes to resolve to 1 when all categories use "none" as placeholder.
        This patches all COCO annotation files in the dataset to use "object" instead.
        """
        import json

        fixed_any = False
        for split in ["train", "valid", "val", "test"]:
            ann_path = os.path.join(dataset_dir, split, "_annotations.coco.json")
            if not os.path.exists(ann_path):
                continue

            with open(ann_path) as f:
                data = json.load(f)

            needs_fix = any(c.get("supercategory") == "none" for c in data.get("categories", []))
            if not needs_fix:
                continue

            for cat in data["categories"]:
                if cat.get("supercategory") == "none":
                    cat["supercategory"] = "object"

            with open(ann_path, "w") as f:
                json.dump(data, f)

            class_names = [c["name"] for c in data["categories"]]
            self.logger.info(f"✅ Fixed supercategories in {split}: {len(class_names)} classes → {class_names}")
            fixed_any = True

        if not fixed_any:
            self.logger.info("ℹ️ No supercategory fix needed — all splits already valid")

TrainerRegistry.register_trainer('RFDETR', RFDETRTrainer)


# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: nedo_vision_training/protos/DatasetService.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'nedo_vision_training/protos/DatasetService.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0nedo_vision_training/protos/DatasetService.proto\x12\x1eSindika.AspNet.App005.Services\"\'\n\x11GetDatasetRequest\x12\x12\n\ndataset_id\x18\x01 \x01(\t\"q\n\x12GetDatasetResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x39\n\x04\x64\x61ta\x18\x03 \x03(\x0b\x32+.Sindika.AspNet.App005.Services.DatasetItem\"l\n\x0b\x44\x61tasetItem\x12\x11\n\tfile_path\x18\x01 \x01(\t\x12J\n\x0b\x61nnotations\x18\x02 \x03(\x0b\x32\x35.Sindika.AspNet.App005.Services.DatasetItemAnnotation\"n\n\x15\x44\x61tasetItemAnnotation\x12\r\n\x05label\x18\x01 \x01(\t\x12\x10\n\x08\x62_box_x1\x18\x02 \x01(\x01\x12\x10\n\x08\x62_box_y1\x18\x03 \x01(\x01\x12\x10\n\x08\x62_box_x2\x18\x04 \x01(\x01\x12\x10\n\x08\x62_box_y2\x18\x05 \x01(\x01\x32\x85\x01\n\x0e\x44\x61tasetService\x12s\n\nGetDataset\x12\x31.Sindika.AspNet.App005.Services.GetDatasetRequest\x1a\x32.Sindika.AspNet.App005.Services.GetDatasetResponseb\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nedo_vision_training.protos.DatasetService_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_GETDATASETREQUEST']._serialized_start=84
  _globals['_GETDATASETREQUEST']._serialized_end=123
  _globals['_GETDATASETRESPONSE']._serialized_start=125
  _globals['_GETDATASETRESPONSE']._serialized_end=238
  _globals['_DATASETITEM']._serialized_start=240
  _globals['_DATASETITEM']._serialized_end=348
  _globals['_DATASETITEMANNOTATION']._serialized_start=350
  _globals['_DATASETITEMANNOTATION']._serialized_end=460
  _globals['_DATASETSERVICE']._serialized_start=463
  _globals['_DATASETSERVICE']._serialized_end=596
# @@protoc_insertion_point(module_scope)



import os

def save_to_file(file_path, data):
    """Save data to a file."""
    with open(file_path, 'w') as file:
        file.write(data)


def load_from_file(file_path):
    """Load data from a file."""
    with open(file_path, 'r') as file:
        return file.read()


def ensure_dir(directory):
    """Ensure that a directory exists."""
    if not os.path.exists(directory):
        os.makedirs(directory)


import argparse
import sys
import signal
import traceback
import time
from .training_service import TrainingService
from .doctor import run_doctor
from .logger.Logger import Logger
from . import __version__


def signal_handler(signum, frame):
    """Handle interrupt signals gracefully."""
    print("\n🛑 Received interrupt signal. Shutting down gracefully...")
    sys.exit(0)


def create_parser():
    """Create and configure the argument parser."""
    parser = argparse.ArgumentParser(
        description="Nedo Vision Training Service Library CLI",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Check system dependencies and requirements
  nedo-training doctor

  # Start training service with token
  nedo-training run --token YOUR_TOKEN

  # Start with custom server configuration
  nedo-training run --token YOUR_TOKEN --server-host custom.server.com --server-port 60000

  # Start with custom REST API port
  nedo-training run --token YOUR_TOKEN --rest-api-port 8081

  # Start with custom intervals
  nedo-training run --token YOUR_TOKEN --system-usage-interval 30 --latency-check-interval 15
        """
    )
    
    parser.add_argument(
        "--version",
        action="version",
        version=f"nedo-vision-training {__version__}"
    )
    
    # Add subcommands
    subparsers = parser.add_subparsers(
        dest='command', 
        help='Available commands',
        required=True
    )
    
    # Doctor command
    subparsers.add_parser(
        'doctor', 
        help='Check system dependencies and requirements',
        description='Run diagnostic checks for CUDA, NVIDIA drivers, and other dependencies'
    )
    
    # Run command
    run_parser = subparsers.add_parser(
        'run',
        help='Start the training service',
        description='Start the Nedo Vision Training Service'
    )
    
    run_parser.add_argument(
        "--token", 
        required=True,
        help="Authentication token provided by the manager"
    )
    
    run_parser.add_argument(
        "--server-host",
        default="localhost",
        help="Server hostname for communication (default: localhost)"
    )
    
    run_parser.add_argument(
        "--server-port",
        type=int,
        default=50051,
        help="Server port for communication (default: 50051)"
    )
    
    run_parser.add_argument(
        "--rest-api-port",
        type=int,
        default=8081,
        help="Manager REST API port (default: 8081)"
    )
    
    run_parser.add_argument(
        "--system-usage-interval",
        type=int,
        default=30,
        help="System usage reporting interval in seconds (default: 30)"
    )
    
    run_parser.add_argument(
        "--latency-check-interval",
        type=int,
        default=10,
        help="Latency monitoring interval in seconds (default: 10)"
    )
    
    return parser


def run_training_service(args):
    """Run the training service with the provided arguments."""
    logger = Logger()
    
    try:
        # Create and configure the training service
        service = TrainingService(
            token=args.token,
            server_host=args.server_host,
            server_port=args.server_port,
            rest_api_port=args.rest_api_port,
            system_usage_interval=args.system_usage_interval,
            latency_check_interval=args.latency_check_interval
        )
        
        # Log startup information
        logger.info("🚀 Starting Nedo Vision Training Service...")
        logger.info(f"🌐 Server: {args.server_host}:{args.server_port}")
        logger.info(f"⏱️ System Usage Interval: {args.system_usage_interval}s")
        logger.info(f"📊 Latency Check Interval: {args.latency_check_interval}s")
        
        # Start the service
        service.run()
        
        # Keep the service running
        try:
            while getattr(service, 'running', False):
                time.sleep(1)
        except KeyboardInterrupt:
            logger.info("\n🛑 Shutdown requested...")
        finally:
            service.stop()
            logger.info("✅ Service stopped successfully")
            
    except Exception as e:
        logger.error(f"❌ Error starting service: {e}")
        traceback.print_exc()
        sys.exit(1)


def main():
    """Main CLI entry point."""
    # Set up signal handlers for graceful shutdown
    signal.signal(signal.SIGINT, signal_handler)
    signal.signal(signal.SIGTERM, signal_handler)
    
    # Parse command line arguments
    parser = create_parser()
    args = parser.parse_args()
    
    # Execute the requested command
    if args.command == 'doctor':
        sys.exit(run_doctor())
    elif args.command == 'run':
        run_training_service(args)


if __name__ == "__main__":
    main()


import os
import importlib
from importlib.util import find_spec
from ...logger.Logger import Logger
from nedo_vision_training.modules.trainer.TrainerRegistry import TrainerRegistry
from nedo_vision_training.modules.algorithm.RFDETR.RFDETRTrainer import RFDETRTrainer

logger = Logger(__name__)

class TrainerFactory:
    _trainers_loaded = False

    @classmethod
    def _load_trainers(cls, algorithm_path='nedo_vision_training.modules.algorithm'):
        """Private method to load trainers dynamically."""
        if cls._trainers_loaded:
            return

        # Locate the base directory of the algorithm module
        spec = find_spec(algorithm_path)
        if not spec or not spec.submodule_search_locations:
            raise ModuleNotFoundError(f"Cannot locate module: {algorithm_path}")

        base_path = spec.submodule_search_locations[0]

        # Walk through the base directory
        for root, _, files in os.walk(base_path):
            for file in files:
                if file.endswith("Trainer.py"):
                    # Construct the module path
                    relative_path = os.path.relpath(root, base_path)
                    module_name = (
                        f"{algorithm_path}.{relative_path.replace(os.sep, '.')}.{file[:-3]}"
                        if relative_path != "."
                        else f"{algorithm_path}.{file[:-3]}"
                    )
                    trainer_class_name = file.replace(".py", "")  # Class name matches the file name

                    try:
                        # Import the module and get the trainer class
                        module = importlib.import_module(module_name)
                        trainer_class = getattr(module, trainer_class_name)
                        TrainerRegistry.register_trainer(trainer_class_name.lower(), trainer_class)

                    except (ImportError, AttributeError) as e:
                        logger.error(f"Error loading trainer {trainer_class_name} from {module_name}: {e}")

        cls._trainers_loaded = True

    @staticmethod
    def get_trainer(algorithm_type):
        """Get a trainer by name, loading trainers if necessary."""
        TrainerFactory._load_trainers()
        return TrainerRegistry.get_trainer(algorithm_type)



import logging

from ..client.TrainingAgentStatusClient import TrainingAgentStatusClient
from ..modules.trainer.Trainer import Trainer
from .DataSenderWorker import DataSenderWorker

logger = logging.getLogger(__name__)

class WorkerManager:
    def __init__(self, config):
        """Initialize all worker threads with the given config (from gRPC)."""
        self.config = config
        self.server_host = self.config.get("server_host")
        self.server_port = self.config.get("server_port")
        self.token = self.config.get("token")

        if not self.server_host:
            raise ValueError("⚠️ Configuration is missing 'server_host'.")
        if not self.server_port:
            raise ValueError("⚠️ Configuration is missing 'server_port'.")
        if not self.token:
            raise ValueError("⚠️ Configuration is missing 'token'.")

        self.status_client = TrainingAgentStatusClient(self.server_host, self.server_port, self.token)
        self.trainer = Trainer(config)
        self.data_sender_worker = DataSenderWorker(config)

    def _start_workers(self):
        """Start processing workers while keeping monitoring workers running."""
        try:
            self.data_sender_worker.start_updating()

            self.trainer.start()
            
            self._update_status("connected")

        except Exception as e:
            logger.error("🚨 Failed to start processing workers.", exc_info=True)

    def _stop_workers(self):
        """Stop processing workers while keeping monitoring workers running."""
        try:
            self.data_sender_worker.stop_updating()

            self.trainer.shutdown_handler()

            self._update_status("disconnected")

        except Exception as e:
            logger.error("🚨 Failed to stop processing workers.", exc_info=True)

    def start_all(self):
        """Start all workers including monitoring workers."""
        try:
            self.data_sender_worker.start()

            self._start_workers()

            logger.info("✅ All workers started successfully.")

        except Exception as e:
            logger.error("🚨 Failed to start all workers.", exc_info=True)

    def stop_all(self):
        """Stop all workers including monitoring workers."""
        try:
            self.data_sender_worker.stop()

            self._stop_workers()

            logger.info("✅ All workers stopped successfully.")

        except Exception as e:
            logger.error("🚨 Failed to stop all workers.", exc_info=True)
    
    def _update_status(self, status_code):
        """
        Update the worker status via gRPC.
        
        Args:
            status_code (str): Status code to report to the server
        """
        try:
            logger.info(f"📡 Updating worker status to {status_code}")
            result = self.status_client.update_status(status_code)
            if result:
                logger.info(f"✅ Status update successful.")
            else:
                logger.warning(f"⚠️ Status update failed.")
        except Exception as e:
            logger.error(f"🚨 Error updating worker status: {str(e)}")



import logging
import grpc
from nedo_vision_training.client.GrpcClientBase import GrpcClientBase
from nedo_vision_training.protos.TrainingJobService_pb2_grpc import TrainingJobServiceStub
from nedo_vision_training.protos.TrainingJobService_pb2 import (
    CreateStatusLogRequest,
    CreateCommandLogRequest,
    CreateMetricsLogRequest,
    UpdateModelFilePathRequest,
    UploadModelRequest,
    UploadModelMetadata
)

logger = logging.getLogger(__name__)

class TrainingLoggerClient(GrpcClientBase):
    def __init__(self, server_host: str, server_port: int = 50051):
        """
        Initialize the training logger client.

        Args:
            server_host (str): The server hostname or IP address.
            server_port (int): The server port. Default is 50051.
        """
        super().__init__(server_host, server_port)

        try:
            self.connect(TrainingJobServiceStub)
        except Exception as e:
            logging.error(f"Failed to connect to gRPC server: {e}")
            self.stub = None

    def upload_model(self, job_id: str, file_path: str) -> dict:
        """
        Upload the trained model file.

        Args:
            job_id (str): The ID of the training job.
            file_path (str): The path to the model file.

        Returns:
            dict: A dictionary containing the result of the upload.
        """
        if not self.stub:
            return {"success": False, "message": "gRPC connection is not established."}

        def request_generator():
            # Send metadata first
            file_name = file_path.split('/')[-1]
            yield UploadModelRequest(metadata=UploadModelMetadata(job_id=job_id, file_name=file_name))

            # Send file chunks
            with open(file_path, 'rb') as f:
                while True:
                    chunk = f.read(1024 * 1024) # 1MB chunks
                    if not chunk:
                        break
                    yield UploadModelRequest(file_chunk=chunk)

        try:
            response = self.stub.UploadModel(request_generator())
            return {
                "success": response.success,
                "message": response.message,
                "file_path": response.file_path
            }
        except grpc.RpcError as e:
            logging.error(f"gRPC error during model upload: {e}")
            return {"success": False, "message": f"gRPC error: {e}"}
        except Exception as e:
            logging.error(f"Unexpected error during model upload: {e}")
            return {"success": False, "message": f"Unexpected error: {e}"}

    def update_status(self, job_id: str, status: str) -> dict:
        """
        Update the training job status.

        Args:
            job_id (str): The ID of the training job.
            status (str): The new status to set.

        Returns:
            dict: A dictionary containing the result of the status update.
        """
        if not self.stub:
            return {"success": False, "message": "gRPC connection is not established."}

        try:
            request = CreateStatusLogRequest(
                job_id=job_id,
                status=status
            )
            
            response = self.handle_rpc(self.stub.CreateStatusLog, request)
            
            if response and response.success:
                return {"success": True, "message": response.message}
            return {"success": False, "message": response.message if response else "Unknown error"}

        except grpc.RpcError as e:
            error_message = getattr(e, "details", lambda: str(e))()
            logger.error(f"gRPC error while updating status: {error_message}")
            return {"success": False, "message": f"RPC error: {error_message}"}

        except Exception as e:
            logger.error(f"Unexpected error while updating status: {str(e)}")
            return {"success": False, "message": "An unexpected error occurred during status update."}

    def log_command(self, job_id: str, command: str) -> dict:
        """
        Log a training command.

        Args:
            job_id (str): The ID of the training job.
            command (str): The command to log.

        Returns:
            dict: A dictionary containing the result of the command logging.
        """
        if not self.stub:
            return {"success": False, "message": "gRPC connection is not established."}

        try:
            request = CreateCommandLogRequest(
                job_id=job_id,
                command=command
            )
            
            response = self.handle_rpc(self.stub.CreateCommandLog, request)
            
            if response and response.success:
                return {"success": True, "message": response.message}
            return {"success": False, "message": response.message if response else "Unknown error"}

        except grpc.RpcError as e:
            error_message = getattr(e, "details", lambda: str(e))()
            logger.error(f"gRPC error while logging command: {error_message}")
            return {"success": False, "message": f"RPC error: {error_message}"}

        except Exception as e:
            logger.error(f"Unexpected error while logging command: {str(e)}")
            return {"success": False, "message": "An unexpected error occurred during command logging."}

    def log_metrics(self, job_id: str, epoch: int, map_50: float, map_50_95: float, 
                   precision: float, recall: float, f1_score: float) -> dict:
        """
        Log training metrics.

        Args:
            job_id (str): The ID of the training job.
            epoch (int): Current training epoch.
            map_50 (float): mAP@0.50 metric.
            map_50_95 (float): mAP@0.50:0.95 metric.
            precision (float): Precision metric.
            recall (float): Recall metric.
            f1_score (float): F1 score metric.

        Returns:
            dict: A dictionary containing the result of the metrics logging.
        """
        if not self.stub:
            return {"success": False, "message": "gRPC connection is not established."}

        try:
            request = CreateMetricsLogRequest(
                job_id=job_id,
                epoch=epoch,
                map_50=map_50,
                map_50_95=map_50_95,
                precision=precision,
                recall=recall,
                f1_score=f1_score
            )
            
            response = self.handle_rpc(self.stub.CreateMetricsLog, request)
            
            if response and response.success:
                return {"success": True, "message": response.message}
            return {"success": False, "message": response.message if response else "Unknown error"}

        except grpc.RpcError as e:
            error_message = getattr(e, "details", lambda: str(e))()
            logger.error(f"gRPC error while logging metrics: {error_message}")
            return {"success": False, "message": f"RPC error: {error_message}"}

        except Exception as e:
            logger.error(f"Unexpected error while logging metrics: {str(e)}")
            return {"success": False, "message": "An unexpected error occurred during metrics logging."}

    def update_model_file_path(self, job_id: str, model_file_path: str) -> dict:
        """
        Update the model file path for a training job.

        Args:
            job_id (str): The ID of the training job.
            model_file_path (str): The path to the resulting model file.

        Returns:
            dict: A dictionary containing the result of the model file path update.
        """
        if not self.stub:
            return {"success": False, "message": "gRPC connection is not established."}

        try:
            request = UpdateModelFilePathRequest(
                job_id=job_id,
                model_file_path=model_file_path
            )
            
            response = self.handle_rpc(self.stub.UpdateModelFilePath, request)
            
            if response and response.success:
                return {"success": True, "message": response.message}
            return {"success": False, "message": response.message if response else "Unknown error"}

        except grpc.RpcError as e:
            error_message = getattr(e, "details", lambda: str(e))()
            logger.error(f"gRPC error while updating model file path: {error_message}")
            return {"success": False, "message": f"RPC error: {error_message}"}

        except Exception as e:
            logger.error(f"Unexpected error while updating model file path: {str(e)}")
            return {"success": False, "message": "An unexpected error occurred during model file path update."}


import threading
import time

from ..logger.Logger import Logger
from ..modules.data_sync.SystemUsageManager import SystemUsageManager

logger = Logger(__name__)

class DataSenderWorker:
    def __init__(self, config: dict, send_interval=None, update_interval=10):
        """
        Initializes the Data Sender Worker.

        Args:
            config (dict): Configuration dictionary.
            send_interval (int): Interval (in seconds) for sending system usage & images. 
                               If None, uses 'system_usage_interval' from config (default: 5).
            update_interval (int): Interval (in seconds) for updating worker sources.
        """
        if not isinstance(config, dict):
            raise ValueError("⚠️ config must be a dictionary.")

        self.config = config
        self.server_host = self.config.get("server_host")
        self.server_port = self.config.get("server_port")
        self.token = self.config.get("token")

        if not self.server_host:
            raise ValueError("⚠️ Configuration is missing 'server_host'.")
        if not self.server_port:
            raise ValueError("⚠️ Configuration is missing 'server_port'.")
        if not self.token:
            raise ValueError("⚠️ Configuration is missing 'token'.")
        
        self.should_update = True

        # Use configurable interval from config, fallback to send_interval parameter or default
        self.send_interval = send_interval if send_interval is not None else self.config.get("system_usage_interval", 5)
        self.update_interval = update_interval

        self.main_thread = None
        self.stop_event = threading.Event()
        self.lock = threading.Lock()

        # Initialize services - pass latency interval to SystemUsageManager
        latency_interval = self.config.get("latency_check_interval", 10)
        self.system_usage_manager = SystemUsageManager(
            self.server_host, 
            self.server_port, 
            self.token, 
            latency_interval=latency_interval
        )

    def start(self):
        """Start the Data Sender Worker threads."""
        with self.lock:
            if self.main_thread and self.main_thread.is_alive():
                logger.warning("⚠️ Data Sender Worker is already running.")
                return

            self.stop_event.clear()

            # ✅ Start the main worker thread (System usage + Image upload)
            self.main_thread = threading.Thread(target=self._run_main_worker, daemon=True)
            self.main_thread.start()

            logger.info(f"🚀 Data Sender Worker started.")

    def stop(self):
        """Stop the Data Sender Worker and Worker Source Updater threads."""
        with self.lock:
            if not self.main_thread or not self.main_thread.is_alive():
                logger.warning("⚠️ Data Sender Worker is not running.")
                return

            self.stop_event.set()

            # ✅ Stop the main worker thread
            if self.main_thread:
                self.main_thread.join(timeout=5)

            self.main_thread = None

            logger.info(f"🛑 Data Sender Worker stopped.")

    def start_updating(self):
        """Start updating worker sources."""
        self.should_update = True

    def stop_updating(self):
        """Stop updating worker sources."""
        self.should_update = False

    def _run_main_worker(self):
        """Main loop for sending system usage and uploading images."""
        try:
            while not self.stop_event.is_set():
                self.system_usage_manager.process_system_usage()
                time.sleep(self.send_interval)
        except Exception as e:
            logger.error("🚨 Unexpected error in main worker loop.")



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: nedo_vision_training/protos/TrainingJobService.proto
# Protobuf Python Version: 6.31.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    6,
    31,
    1,
    '',
    'nedo_vision_training/protos/TrainingJobService.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n4nedo_vision_training/protos/TrainingJobService.proto\x12\x1eSindika.AspNet.App005.Services\"8\n\x16\x43reateStatusLogRequest\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\";\n\x17\x43reateStatusLogResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\":\n\x17\x43reateCommandLogRequest\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12\x0f\n\x07\x63ommand\x18\x02 \x01(\t\"<\n\x18\x43reateCommandLogResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"\x90\x01\n\x17\x43reateMetricsLogRequest\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12\r\n\x05\x65poch\x18\x02 \x01(\x04\x12\x0e\n\x06map_50\x18\x03 \x01(\x01\x12\x11\n\tmap_50_95\x18\x04 \x01(\x01\x12\x11\n\tprecision\x18\x05 \x01(\x01\x12\x0e\n\x06recall\x18\x06 \x01(\x01\x12\x10\n\x08\x66\x31_score\x18\x07 \x01(\x01\"<\n\x18\x43reateMetricsLogResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"E\n\x1aUpdateModelFilePathRequest\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12\x17\n\x0fmodel_file_path\x18\x02 \x01(\t\"?\n\x1bUpdateModelFilePathResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\"{\n\x12UploadModelRequest\x12G\n\x08metadata\x18\x01 \x01(\x0b\x32\x33.Sindika.AspNet.App005.Services.UploadModelMetadataH\x00\x12\x14\n\nfile_chunk\x18\x02 \x01(\x0cH\x00\x42\x06\n\x04\x64\x61ta\"8\n\x13UploadModelMetadata\x12\x0e\n\x06job_id\x18\x01 \x01(\t\x12\x11\n\tfile_name\x18\x02 \x01(\t\"J\n\x13UploadModelResponse\x12\x0f\n\x07success\x18\x01 \x01(\x08\x12\x0f\n\x07message\x18\x02 \x01(\t\x12\x11\n\tfile_path\x18\x03 \x01(\t2\xb4\x05\n\x12TrainingJobService\x12\x82\x01\n\x0f\x43reateStatusLog\x12\x36.Sindika.AspNet.App005.Services.CreateStatusLogRequest\x1a\x37.Sindika.AspNet.App005.Services.CreateStatusLogResponse\x12\x85\x01\n\x10\x43reateCommandLog\x12\x37.Sindika.AspNet.App005.Services.CreateCommandLogRequest\x1a\x38.Sindika.AspNet.App005.Services.CreateCommandLogResponse\x12\x85\x01\n\x10\x43reateMetricsLog\x12\x37.Sindika.AspNet.App005.Services.CreateMetricsLogRequest\x1a\x38.Sindika.AspNet.App005.Services.CreateMetricsLogResponse\x12\x8e\x01\n\x13UpdateModelFilePath\x12:.Sindika.AspNet.App005.Services.UpdateModelFilePathRequest\x1a;.Sindika.AspNet.App005.Services.UpdateModelFilePathResponse\x12x\n\x0bUploadModel\x12\x32.Sindika.AspNet.App005.Services.UploadModelRequest\x1a\x33.Sindika.AspNet.App005.Services.UploadModelResponse(\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'nedo_vision_training.protos.TrainingJobService_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_CREATESTATUSLOGREQUEST']._serialized_start=88
  _globals['_CREATESTATUSLOGREQUEST']._serialized_end=144
  _globals['_CREATESTATUSLOGRESPONSE']._serialized_start=146
  _globals['_CREATESTATUSLOGRESPONSE']._serialized_end=205
  _globals['_CREATECOMMANDLOGREQUEST']._serialized_start=207
  _globals['_CREATECOMMANDLOGREQUEST']._serialized_end=265
  _globals['_CREATECOMMANDLOGRESPONSE']._serialized_start=267
  _globals['_CREATECOMMANDLOGRESPONSE']._serialized_end=327
  _globals['_CREATEMETRICSLOGREQUEST']._serialized_start=330
  _globals['_CREATEMETRICSLOGREQUEST']._serialized_end=474
  _globals['_CREATEMETRICSLOGRESPONSE']._serialized_start=476
  _globals['_CREATEMETRICSLOGRESPONSE']._serialized_end=536
  _globals['_UPDATEMODELFILEPATHREQUEST']._serialized_start=538
  _globals['_UPDATEMODELFILEPATHREQUEST']._serialized_end=607
  _globals['_UPDATEMODELFILEPATHRESPONSE']._serialized_start=609
  _globals['_UPDATEMODELFILEPATHRESPONSE']._serialized_end=672
  _globals['_UPLOADMODELREQUEST']._serialized_start=674
  _globals['_UPLOADMODELREQUEST']._serialized_end=797
  _globals['_UPLOADMODELMETADATA']._serialized_start=799
  _globals['_UPLOADMODELMETADATA']._serialized_end=855
  _globals['_UPLOADMODELRESPONSE']._serialized_start=857
  _globals['_UPLOADMODELRESPONSE']._serialized_end=931
  _globals['_TRAININGJOBSERVICE']._serialized_start=934
  _globals['_TRAININGJOBSERVICE']._serialized_end=1626
# @@protoc_insertion_point(module_scope)



"""
Nedo Vision Training Service Library

A comprehensive library for training AI models in the Nedo Vision platform.
This library provides tools and services for training computer vision models,
managing training workflows, and integrating with the broader Nedo Vision ecosystem.

Features:
- AI model training orchestration
- gRPC-based service communication
- AWS integration for cloud resources
- GPU monitoring and management
- Flexible training pipeline configuration

Example:
    >>> from nedo_vision_training import TrainingService
    >>> service = TrainingService()
    >>> # Configure and start training...
"""

from .training_service import TrainingService
from .exceptions import TrainingServiceError, ConfigurationError

__version__ = "1.3.4"
__author__ = "Willy Achmat Fauzi"
__email__ = "willy.achmat@gmail.com"
__license__ = "MIT"
__title__ = "nedo-vision-training"
__description__ = "A comprehensive training service library for AI models in the Nedo Vision platform"
__url__ = "https://gitlab.com/sindika/research/nedo-vision/nedo-vision-training-service"

__all__ = [
    "TrainingService", 
    "TrainingServiceError", 
    "ConfigurationError",
    "__version__",
    "__author__",
    "__email__",
    "__license__",
    "__title__",
    "__description__",
    "__url__",
] 